"""Classes common to several modules"""
#TODO: elaborate module doc
from __future__ import print_function, absolute_import, division

from collections import defaultdict

import natsort


try:
    from itertools import izip
except ImportError:  #python3.x
    izip = zip

class KatanaException(Exception):
    """Flagging cases that we can not process at this time."""
    def __init__(self, msg, *args):
        #pylint: disable=star-args
        error_msg = msg.format(*[str(i) for i in args])
        super(KatanaException, self).__init__(error_msg)


class PrimerStats(object):
    '''Collects simple counts for overall reads and reads per primer.'''
    STAT_KEYS = ["chrom", "target_id", "sense_start", "sense_count",
                 "antisense_count", "sense_percent"]
    def __init__(self):
        self.total_read_count = 0
        self._primer_stats = defaultdict(int)


    def _sense_percent(self, primer):
        sense_read_count = self._primer_stats[primer, True]
        antisense_read_count = self._primer_stats[primer, False]
        return int(100 * sense_read_count / \
                   (antisense_read_count + sense_read_count))

    def add_read_primer(self, read, primer_pair):
        self.total_read_count +=1
        stat_key = (primer_pair, read.is_positive_strand)
        self._primer_stats[stat_key] += 1

    @property
    def primer_pairs(self):
        sort_key = lambda x:(x.chrom, x.sense_start, x.target_id)
        primers = set([primer for (primer, _) in self._primer_stats])
        return natsort.natsorted(primers, key=sort_key)

    def stats(self, primer_pair):
        values = [primer_pair.chrom,
                  primer_pair.target_id,
                  primer_pair.sense_start,
                  self._primer_stats[primer_pair, True],
                  self._primer_stats[primer_pair, False],
                  self._sense_percent(primer_pair)]
        return dict(izip(PrimerStats.STAT_KEYS, values))


#TODO: Extend log method to emit subset to screen and full set to file
class PrimerStatsDumper(object):
    '''Prints PrimerStats object to log.'''
    def __init__(self, log_method):
        self._log_method = log_method

    def dump(self, primer_stats):
        # pylint: disable=star-args
        stat_format = "PRIMER_STATS" + "|{}" * len(primer_stats.STAT_KEYS)
        self._log_method(stat_format.format(*primer_stats.STAT_KEYS))
        for primer_pair in primer_stats.primer_pairs:
            stat_dict = primer_stats.stats(primer_pair)
            stats = [stat_dict[key] for key in primer_stats.STAT_KEYS]
            self._log_method(stat_format.format(*stats))

#TODO: sort methods
class Read(object):
    '''Lightweight wrapper around AlignedSegment'''
    def __init__(self, aligned_segment, input_bamfile):
        self.aligned_segment = aligned_segment
        self.input_bamfile = input_bamfile

    @property
    def is_positive_strand(self):
        return not self.aligned_segment.is_reverse

    def _key(self, mate=False):
        key = None
        if not mate:
            key= (self.aligned_segment.query_name,
                  self.is_positive_strand,
                  self.input_bamfile.getrname(self.aligned_segment.reference_id),
                  self.aligned_segment.reference_start,
                  self.aligned_segment.is_read1)
        elif self.aligned_segment.is_paired and \
                (not self.aligned_segment.mate_is_unmapped):
            key = (self.aligned_segment.query_name,
                   not self.aligned_segment.mate_is_reverse,
                   self.input_bamfile.getrname(self.aligned_segment.next_reference_id),
                   self.aligned_segment.next_reference_start,
                   not self.aligned_segment.is_read1)
        return key

    @property
    def key(self):
        return self._key(False)

    @property
    def next_reference_start(self):
        return self.aligned_segment.next_reference_start

    @next_reference_start.setter
    def next_reference_start(self, value):
        self.aligned_segment.next_reference_start = value


    @property
    def mate_key(self):
        return self._key(True)

    @property
    def is_unmapped(self):
        return self.aligned_segment.is_unmapped

    @property
    def mate_is_paired(self):
        return self.aligned_segment.is_paired

    @mate_is_paired.setter
    def is_paired(self, value):
        self.aligned_segment.is_paired = value
        if not value:
            self.aligned_segment.mate_is_unmapped = False
            self.aligned_segment.mate_is_reverse = False
            self.aligned_segment.is_proper_pair = False
            self.aligned_segment.is_read1 = False
            self.aligned_segment.is_read2 = False
            self.aligned_segment.next_reference_id = -1
            self.aligned_segment.next_reference_start = 0

    @property
    def query_name(self):
        return self.aligned_segment.query_name

    @property
    def reference_name(self):
        return self.input_bamfile.getrname(self.aligned_segment.reference_id)
#        return self.aligned_segment.rname

    @property
    def reference_start(self):
        return self.aligned_segment.reference_start

    @reference_start.setter
    def reference_start(self, value):
        self.aligned_segment.reference_start = value

    @property
    def reference_end(self):
        return self.aligned_segment.reference_end

    @property
    def cigarstring(self):
        return self.aligned_segment.cigarstring

    @cigarstring.setter
    def cigarstring(self, value):
        self.aligned_segment.cigarstring = value

    def set_tag(self, tag_name, tag_value, tag_type):
        self.aligned_segment.set_tag(tag_name, tag_value, tag_type)

    @staticmethod
    def iter(aligned_segment_iter, input_bamfile):
        for aligned_segment in aligned_segment_iter:
            yield Read(aligned_segment, input_bamfile)

class _NullPrimerPair(object):
    #pylint: disable=too-few-public-methods
    def __init__(self):
        self.target_id = "PRIMER_NOT_RECOGNIZED"
        self.chrom = "."
        self.sense_start = 0

    @staticmethod
    def softclip_primers(old_cigar):
        return old_cigar

    @property
    def is_unmatched(self):
        #pylint: disable=no-self-use
        return True

class PrimerPair(object):
    '''Sense start and antisense regions should be start and end (inclusive and
    exclusive respectively), zero-based, positive genomic index.'''
    _all_primers = {}
    NULL = _NullPrimerPair()

    def __init__(self,
                 target_id,
                 chrom,
                 sense_primer_region,
                 antisense_primer_region):
        self.target_id = target_id
        self.chrom = chrom
        self.sense_start = sense_primer_region[0]
        self._query_region_start = sense_primer_region[1]
        self._query_region_end =  antisense_primer_region[0]
        self._add_primer(chrom,
                         sense_primer_region[0],
                         antisense_primer_region[1])

    def _add_primer(self, chrom, sense_start, antisense_start):
        sense_key = (chrom, sense_start, True)
        antisense_key = (chrom, antisense_start, False)
        self._all_primers[sense_key] = self
        self._all_primers[antisense_key] = self

    @staticmethod
    def _key_for_read(read):
        if read.is_positive_strand:
            return (read.reference_name, read.reference_start, True)
        else:
            return (read.reference_name, read.reference_end, False)

    @staticmethod
    def get_primer_pair(read):
        '''Matches a read to a primer pair based on exact start position of the
        read; returns NullPrimerPair if no match.

        Note this exact approach will disqualify some otherwise valid reads,
        e.g. if a read adapter is not trimmed precisely, those reads will have
        an "early start" and will not be correctly matched with their primer
        pair.
        '''
        try:
            read_key = PrimerPair._key_for_read(read)
            primer_pair = PrimerPair._all_primers[read_key]
            return primer_pair
        except KeyError:
            return PrimerPair.NULL

    @property
    def is_unmatched(self):
        #pylint: disable=no-self-use
        return False

    def softclip_primers(self, old_cigar):
        return old_cigar.softclip_target(self._query_region_start,
                                         self._query_region_end)

class _NullReadTransformation(object):
    def __init__(self):
        self.is_paired = False
        self.is_unmapped = False
        self.primer_pair = PrimerPair.NULL
        self.reference_start = 0
        self.cigar = ""
        self.is_cigar_valid = False
        self.filters=("NULL",)


class ReadTransformation(object):
    '''Lightweight container for changes to be made to a read.'''

    NULL = _NullReadTransformation()

    def __init__(self,
                 read,
                 primer_pair,
                 new_cigar,
                 filter_builder=None):
        self.is_paired = read.is_paired
        self.is_unmapped = read.is_unmapped
        self.primer_pair = primer_pair
        self.reference_start = new_cigar.reference_start
        self.cigar = new_cigar.cigar
        self.is_cigar_valid = new_cigar.is_valid
        if filter_builder == None:
            self.filters = ()
        else:
            self.filters = tuple(filter_builder(self))

    def __eq__(self, other):
        return self.__dict__ == other.__dict__

    def __repr__(self):
        return ("{}("
                "primer_pair={}, "
                "reference_start={}, "
                "cigar='{}', "
                "is_paired={}, "
                "filters={})").format(self.__class__,
                                      self.primer_pair.target_id,
                                      self.reference_start,
                                      self.cigar,
                                      self.is_paired,
                                      self.filters)




#! /usr/bin/env python
"""Softclips primers at edge of aligned reads based on primer locations; emits
new BAM file with clipped reads optionally excluding alignments that did not
match primer locations."""
#TODO: elaborate module doc
#TODO: see TODO.rst
#TODO: Add to PyPI

##   Copyright 2014 Bioinformatics Core, University of Michigan
##
##   Licensed under the Apache License, Version 2.0 (the "License");
##   you may not use this file except in compliance with the License.
##   You may obtain a copy of the License at
##
##       http://www.apache.org/licenses/LICENSE-2.0
##
##   Unless required by applicable law or agreed to in writing, software
##   distributed under the License is distributed on an "AS IS" BASIS,
##   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
##   See the License for the specific language governing permissions and
##   limitations under the License.

from __future__ import print_function, absolute_import, division

import argparse
import csv
from datetime import datetime
import resource
import sys
import time
import traceback

import pysam

import katana
import katana.cigar as cigar
import katana.readhandler as readhandler
from katana.util import KatanaException, PrimerStats, PrimerStatsDumper, \
    PrimerPair, Read, ReadTransformation


__version__ = katana.__version__

DESCRIPTION=\
'''Match each alignment in input BAM to primer, softclipping the primer region.

Katana matches each read to its corresponding primer pair based on start
position of the read. Katana then soft-clips the primer region from the edge of
the read sequence, rescuing the signal of true variants measured by overlapping
amplicons. The output is conceptually similar to hard-clipping the primers from
the original FASTQ reads based on sequence identity but with the advantage that
retaining the primers during alignment improves alignment quality.
'''

class _KatanaUsageError(Exception):
    """Raised for malformed command or invalid arguments."""
    def __init__(self, msg, *args):
        super(_KatanaUsageError, self).__init__(msg, *args)


class _KatanaArgumentParser(argparse.ArgumentParser):
    """Argument parser that raises UsageError instead of exiting."""
    #pylint: disable=too-few-public-methods
    def error(self, message):
        '''Suppress default exit behavior'''
        raise _KatanaUsageError(message)


#TODO: make this a logger object that writes to file and console and
# supports debug and info calls
def _log(msg_format, *args):
    timestamp = datetime.now().strftime('%Y/%m/%d %H:%M:%S')
    try:
        print("{}|{}".format(timestamp, msg_format).format(*args),
              file=sys.stderr)
    except IndexError:
        print(args)
    sys.stderr.flush()

def _filter_builder(read_transformation):
    filters = []
    if read_transformation.is_unmapped:
        filters.append("UNMAPPED_ALIGNMENT")
    else:
        if read_transformation.primer_pair.is_unmatched:
            filters.append("UNMATCHED_PRIMER_PAIR")
        if not read_transformation.is_cigar_valid:
            filters.append("INVALID_CIGAR")
    return filters

#TODO: refactor to expedite testing (e.g. clipped_cigar_provider,
#  cached_clipped_cigar_provider)
def _build_read_transformations(read_iter, filter_builder):
    read_transformations = {}
    read_count = 0
    cigar_cache={}
    for read in read_iter:
        try:
            read_count += 1
            primer_pair = PrimerPair.get_primer_pair(read)
            key = (primer_pair, read.reference_start, read.cigarstring)
            if not cigar_cache.get(key):
                old_cigar = cigar.cigar_factory(read)
                new_cigar = primer_pair.softclip_primers(old_cigar)
                cigar_cache[key] = new_cigar
            new_cigar = cigar_cache[key]
            transform = ReadTransformation(read,
                                           primer_pair,
                                           new_cigar,
                                           filter_builder)
            read_transformations[read.key] = transform
        except Exception as exception:
            msg = "Problem with read {} [line {}] and primer pair {}: {}"
            raise KatanaException(msg.format(read.query_name,
                                             read_count,
                                             primer_pair.target_id,
                                             exception))
    _log("Built transforms for [{}] alignments", read_count)
    return read_transformations


def _handle_reads(read_handlers, read_iter, read_transformations):
    for handler in read_handlers:
        handler.begin()
    for read in read_iter:
        read_transformation = read_transformations[read.key]
        mate_transformation = read_transformations.get(read.mate_key,
                                                       ReadTransformation.NULL)
        try:
            for handler in read_handlers:
                handler.handle(read, read_transformation, mate_transformation)
        except StopIteration:
            pass
    for handler in read_handlers:
        handler.end()

def _initialize_primer_pairs(base_reader):
    dict_reader = csv.DictReader(base_reader, delimiter='\t')
    for row in dict_reader:
        sense_start = int(row["Sense Start"]) - 1
        sense_end = sense_start + len(row["Sense Sequence"])
        antisense_start = int(row["Antisense Start"])
        antisense_end = antisense_start - len(row["Antisense Sequence"])
        PrimerPair(row["Customer TargetID"],
                   "chr" + row["Chr"],  #TODO: this prefix seems hackish?
                   (sense_start, sense_end),
                   (antisense_end, antisense_start))

def _build_handlers(input_bam_filename,
                    output_bam_filename,
                    include_unmatched_reads):
    stats = readhandler.StatsHandler(PrimerStats(),
                                      PrimerStatsDumper(log_method=_log))
    exclude = readhandler.ExcludeNonMatchedReadHandler(log_method=_log)
    tag = readhandler.AddTagsReadHandler()
    transform = readhandler.TransformReadHandler()
    write = readhandler.WriteReadHandler(input_bam_filename,
                                          output_bam_filename,
                                          log_method=_log)
    handlers = [stats, tag, transform, exclude, write]
    if include_unmatched_reads:
        handlers.remove(exclude)
    return handlers

def _parse_command_line_args(arguments):
    parser = _KatanaArgumentParser( \
        formatter_class=argparse.RawTextHelpFormatter,
        usage="katana primer_manifest input_bam output_bam",
        description=(DESCRIPTION))

    parser.add_argument("-V",
                        "--version",
                        action='version',
                        version=__version__)
    parser.add_argument('primer_manifest',
                        help="path to primer manifest (tab-separated text)")
    parser.add_argument('input_bam',
                        help="path to input BAM")
    parser.add_argument('output_bam',
                        help="path to output BAM")
    parser.add_argument("--preserve_all_alignments",
                        action="store_true",
                        help=("Preserve all incoming alignments (even if they "
                              "are unmapped, cannot be matched with primers, "
                              "result in invalid CIGARs, etc.)"))
    args = parser.parse_args(arguments)
    return args

def _peak_memory():
    peak_memory = resource.getrusage(resource.RUSAGE_SELF).ru_maxrss
    peak_memory_mb = peak_memory/1024
    if sys.platform == 'darwin':
        peak_memory_mb /= 1024
    return int(peak_memory_mb)


#TODO: test
#TODO: check input files exist
def main(command_line_args=None):
    '''Katana entry point.'''
    try:
        start_time = time.time()
        if not command_line_args:
            command_line_args = sys.argv
        args = _parse_command_line_args(command_line_args[1:])

        _log("Reading primer pairs from [{}]", args.primer_manifest)
        with open(args.primer_manifest, "r") as input_primer_manifest:
            _initialize_primer_pairs(input_primer_manifest)
        _log("Read [{}] primer pairs", len(PrimerPair._all_primers))
        input_bamfile = None

        _log("Building transformations from BAM [{}]", args.input_bam)
        #pylint: disable=no-member
        input_bamfile = pysam.AlignmentFile(args.input_bam,"rb")
        aligned_segment_iter = input_bamfile.fetch()
        read_iter = Read.iter(aligned_segment_iter, input_bamfile)
        read_transformations = _build_read_transformations(read_iter,
                                                            _filter_builder)

        _log("Writing transformed alignments to [{}]", args.output_bam)
        handlers = _build_handlers(args.input_bam,
                                   args.output_bam,
                                   args.preserve_all_alignments)
        aligned_segment_iter = input_bamfile.fetch()
        read_iter = Read.iter(aligned_segment_iter, input_bamfile)
        _handle_reads(handlers, read_iter, read_transformations)

        elapsed_time = int(time.time() - start_time)
        _log("Done ({} seconds, {}mb peak memory)",
             elapsed_time,
             _peak_memory())
    except _KatanaUsageError as usage_error:
        message = "katana usage problem: {}".format(str(usage_error))
        print(message, file=sys.stderr)
        print("See 'katana --help'.", file=sys.stderr)
        sys.exit(1)
    except Exception: #pylint: disable=broad-except
        _log("ERROR: An unexpected error occurred")
        _log(traceback.format_exc())
        exit(1)
    finally:
        try:
            if input_bamfile:
                input_bamfile.close()
        except NameError:
            pass


if __name__ == '__main__':
    #import cProfile
    #cProfile.run('main()')
    main(sys.argv)



"""Basic CIGAR manipulation and querying. """

from __future__ import print_function, absolute_import, division

import itertools
import re

import katana.util as util


class CigarUtil(object):
    _QUERY_CONSUMING = set(list("MIS=X"))
    _REF_CONSUMING = set(list("MDNS=X"))
    _REGEX_CIGAR = re.compile("([0-9]+)([MIDNSHP=X])")
    _REGEX_MATCHING_OP = re.compile("[MX=]")
    _REGEX_NON_HARDCLIP = re.compile("[^H]")
    _REGEX_REF_CONSUMING = re.compile("[MDNS=X]")
    _REGEX_REQUIRED_OPS = re.compile("[MIDN=X]")
    _REGEX_QUERY_CONSUMING = re.compile("[MIS=X]")
    _REGEX_QUERY_NON_CONSUMING = re.compile("[DNP]")

    def __init__(self, reference_start, cigar=None, cigar_profile=None):
        self.reference_start = reference_start
        self.cigar = ""
        self.cigar_profile = ""
        if cigar:
            self.cigar = cigar
        elif cigar_profile:
            self.cigar = self._collapse_cigar_profile(cigar_profile)
        if cigar_profile:
            self.cigar_profile = cigar_profile
        else:
            self.cigar_profile = self._expand_cigar(self.cigar)
        self.query_length = \
                len(self._REGEX_QUERY_CONSUMING.findall(self.cigar_profile))
        self.is_valid = self._REGEX_REQUIRED_OPS.search(self.cigar) is not None

    def __repr__(self):
        return ("{}(reference_start={}, "
                "cigar='{}')").format(self.__class__,
                                      self.reference_start,
                                      self.cigar)

    def __eq__(self, other):
        return self.__dict__ == other.__dict__

    def _is_ref_consuming(self, cigar_op):
        return cigar_op in self._REF_CONSUMING

    def _is_match(self, cigar_op):
        return self._REGEX_MATCHING_OP.search(cigar_op) != None

    def _softclip(self, cigar_profile):
        profile = self._REGEX_QUERY_NON_CONSUMING.sub("", cigar_profile)
        return self._REGEX_NON_HARDCLIP.sub("S", profile)

    def _cut_at_first_match(self, profile):
        cut_index = 0
        first_match = self._REGEX_MATCHING_OP.search(profile)
        if first_match:
            cut_index = first_match.start(0)
        return (profile[0:cut_index], profile[cut_index:])

    def _softclip_to_first_match(self, old_pos, old_profile):
        new_pos = old_pos
        (pre_match, post_match) = self._cut_at_first_match(old_profile)
        new_pos += len(re.findall(self._REGEX_REF_CONSUMING, pre_match))
        pre_match = self._softclip(pre_match)
        return (new_pos, pre_match + post_match)

    def _expand_cigar(self, cigar_string):
        expanded_cigar = []
        for cigar_tuple in self._REGEX_CIGAR.findall(cigar_string):
            expanded_cigar.append(int(cigar_tuple[0]) * cigar_tuple[1])
        return "".join(expanded_cigar)

    @staticmethod
    def _collapse_cigar_profile(profile):
        op_strings = ["".join(g) for _, g in itertools.groupby(profile)]
        new_cigar = [str(len(op)) + op[0] for op in op_strings]
        return "".join(new_cigar)

    def _pos_profiles(self, profile):
        '''Returns a list of tuple of first match index and a list of profiles.
        Each list element is the cigar profile for that reference position; i.e.
        the length of this list is the length of consumed reference bases.
        '''
        pos_profiles = list()
        pos = 0
        pos_profiles.append([])
        first_match_index = -1
        for cigar_op in profile:
            if pos >= len(pos_profiles):
                pos_profiles.append([])
            pos_profiles[pos].append(cigar_op)
            if self._is_match(cigar_op) and first_match_index == -1:
                first_match_index = pos
            if self._is_ref_consuming(cigar_op):
                pos += 1

        profiles = ["".join(x) for x in pos_profiles]

        return (first_match_index, profiles)

    def _partition_cigar(self, region_start, region_end):
        '''Split the profile into 3-tuple: before region, in region,
        and after region; each tuple defines the start coordinate and
        profile fragment.

        ref_start, region start, and region_end are genomic coordinates.

        Regions can extend outside the range of the read but the resulting
        start coordinates are constrained to the range of the read.  For
        example, if region_end overhangs the end of the read, the "after region"
        start coordinate will be "pushed back" to the trailing edge of the
        read.'''
        ref_start = self.reference_start
        profile = self.cigar_profile
        (first_match_index, pos_profile) = self._pos_profiles(profile)

        read_start = ref_start - first_match_index
        read_end = read_start + len(pos_profile)
        constrain = lambda x: max(read_start, min(read_end, x))

        region_start_index = max(region_start - read_start, 0)
        region_end_index = region_end - read_start

        before_profile = "".join(pos_profile[0:region_start_index])
        region_profile="".join(pos_profile[region_start_index:region_end_index])
        after_profile="".join(pos_profile[region_end_index:])

        return (CigarUtil(read_start,
                          cigar_profile=before_profile),
                CigarUtil(constrain(region_start),
                          cigar_profile=region_profile),
                CigarUtil(constrain(region_end),
                          cigar_profile=after_profile))

    def _assert_query_lengths_match(self, new_cigar):
        if self.query_length != new_cigar.query_length:
            msg = ("Old CIGAR query length [{}] ({}) != new CIGAR length"
                    "[{}] ({})").format(self.cigar,
                                        self.query_length,
                                        new_cigar.cigar,
                                        new_cigar.query_length)
            raise util.KatanaException(msg)

    def softclip_target(self, target_start, target_end):
        (pre_target, target, post_target) = self._partition_cigar(target_start,
                                                                  target_end)
        pre_profile = self._softclip(pre_target.cigar_profile)
        (new_pos,
         target_profile) = self._softclip_to_first_match(target.reference_start,
                                                         target.cigar_profile)
        post_profile = self._softclip(post_target.cigar_profile)
        new_profile = pre_profile + target_profile + post_profile
        new_cigar = CigarUtil(new_pos, cigar_profile = new_profile)
        self._assert_query_lengths_match(new_cigar)
        return  new_cigar


class NullCigarUtil(object):
#pylint: disable=unused-argument
    def __init__(self, reference_start):
        self.reference_start = reference_start
        self.cigar = "*"
        self.is_valid = True
        self.query_length = 0

    def softclip_target(self, target_start, target_end):
        return self


def cigar_factory(read):
    if not read.cigarstring or read.cigarstring == "*":
        return NullCigarUtil(read.reference_start)
    else:
        return CigarUtil(read.reference_start, read.cigarstring)




#pylint: disable=line-too-long
import os

from setuptools import find_packages, setup

import katana


def read(*paths):
    """Build a file path from *paths* and return the contents."""
    with open(os.path.join(*paths), 'r') as filename:
        return filename.read()

setup(name='Katana',
      version = katana.__version__,
      description=('Command-line tool to soft-clip reads based on primer locations.'),
      long_description=(read('README.rst') + '\n\n' +
                        read('CHANGELOG.rst') + '\n\n' +
                        read('AUTHORS.rst')),
      url='https://github.com/umich-brcf-bioinf/Katana',
      author='University of Michigan Bioinformatics Core',
      author_email='bfx-katana@umich.edu',
      license='Apache',
      packages=find_packages(exclude=['test*']),
      classifiers=['Development Status :: 4 - Beta',
                   'Environment :: Console',
                   'Intended Audience :: Science/Research',
                   'License :: OSI Approved :: Apache Software License',
                   'Operating System :: Unix',
                   'Programming Language :: Python :: 2.7',
                   'Programming Language :: Python :: 3',
                   'Topic :: Scientific/Engineering :: Bio-Informatics'],
      keywords='bioinformatic exome-seq DNA-seq BAM',
      install_requires=['pysam', 'natsort'],
      entry_points={'console_scripts': ['katana=katana.clipper:main']},
      test_suite='nose.collector',
      tests_require=['nose', 'pysam', 'natsort','testfixtures'],
      zip_safe=False)



"""Various classes that handle reads"""
#TODO: elaborate module doc
from __future__ import print_function, absolute_import, division

from collections import defaultdict
import os

import pysam


# I would rather just say pysam.index(...), but since that global is
# added dynamically, Eclipse flags this as a compilation problem. So
# instead we connect directly to the pysam.SamtoolsDispatcher.
def pysam_index(input_filename):
    pysam.SamtoolsDispatcher("index", None).__call__(input_filename,
                                                     catch_stdout=False)

def pysam_sort(input_filename, output_prefix):
    pysam.SamtoolsDispatcher("sort", None).__call__(input_filename,
                                                    output_prefix,
                                                    catch_stdout=False)


class _BaseReadHandler(object):
    def begin(self):
        pass
    def handle(self, read, read_transformation, mate_transformation):
        pass
    def end(self):
        pass


class AddTagsReadHandler(_BaseReadHandler):
    '''Adds original read values and other explanatory tags.'''
    def handle(self, read, read_transformation, mate_transformation):
        primer_pair = read_transformation.primer_pair
        read.set_tag("X0", primer_pair.target_id, "Z")
        read.set_tag("X1", read.cigarstring, "Z")
        read.set_tag("X2", read.reference_start, "i")
        read.set_tag("X3", read.reference_end, "i")
        if read_transformation.filters:
            filter_string = ",".join(read_transformation.filters)
            read.set_tag("X4", filter_string, "Z")


class ExcludeNonMatchedReadHandler(_BaseReadHandler):
    '''Excludes reads from further processing'''
    _STOP_ITERATION_EXCEPTION = StopIteration()
    def __init__(self, log_method):
        self._log_method = log_method
        self._all_exclusions = defaultdict(int)

    def handle(self, read, read_transformation, mate_transformation):
        if mate_transformation.filters:
            read.is_paired = False
        if read_transformation.filters:
            self._all_exclusions[read_transformation.filters] += 1
            raise self._STOP_ITERATION_EXCEPTION

    def end(self):
        for (filters, count) in self._all_exclusions.items():
            msg = "EXCLUDE|{} alignments were excluded because: {}"
            self._log_method(msg, count, ",".join(filters))


class StatsHandler(_BaseReadHandler):
    '''Processes reads and primers connecting PrimerStats and
    PrimerStatsDumper'''
    def __init__(self, primer_stats, primer_stats_dumper):
        self._primer_stats = primer_stats
        self._primer_stats_dumper = primer_stats_dumper

    def handle(self, read, read_transformation, mate_transformation):
        self._primer_stats.add_read_primer(read,
                                           read_transformation.primer_pair)

    def end(self):
        self._primer_stats_dumper.dump(self._primer_stats)


class TransformReadHandler(_BaseReadHandler):
    '''Updates reference_start, cigar string, and mate_reference_start.'''
    def handle(self, read, read_transformation, mate_transformation):
        read.reference_start = read_transformation.reference_start
        read.cigarstring = read_transformation.cigar
        if read.is_paired and not mate_transformation.is_unmapped:
            read.next_reference_start = mate_transformation.reference_start


#TODO: Add PG and CO header lines for tags
class WriteReadHandler(_BaseReadHandler):
    '''Writes reads to a BAM file (ultimately sorting and indexing the BAM).'''
    def __init__(self,
                 input_bam_filename,
                 output_bam_filename,
                 log_method):
        self._input_bam_filename = input_bam_filename
        output_dir = os.path.dirname(output_bam_filename)
        output_basename = "tmp_unsorted_" \
                + os.path.basename(output_bam_filename)
        self._tmp_bam_filename = os.path.join(output_dir, output_basename)
        self._output_bam_filename = output_bam_filename
        self._bamfile = None
        self._log = log_method
        self._read_count = 0

    def begin(self):
        #pylint: disable=no-member
        input_bam = None
        try:
            input_bam = pysam.AlignmentFile(self._input_bam_filename, "rb")
            self._bamfile = pysam.AlignmentFile(self._tmp_bam_filename,
                                                "wb",
                                                template=input_bam)
        finally:
            if input_bam:
                input_bam.close()

    def handle(self, read, read_transformation, mate_transformation):
        self._read_count += 1
        self._bamfile.write(read.aligned_segment)

    def end(self):
        self._bamfile.close()
        self._bamfile = None
        output_root = os.path.splitext(self._output_bam_filename)[0]
        self._log("WRITE_BAM|sorting BAM")
        pysam_sort(self._tmp_bam_filename, output_root)
        self._log("WRITE_BAM|indexing BAM")
        pysam_index(self._output_bam_filename)
        os.remove(self._tmp_bam_filename)
        self._log("WRITE_BAM|wrote [{}] alignments to [{}]",
                  self._read_count,
                  self._output_bam_filename)
